#include "stdafx.h"
#include "resource.h"

// I got the idea of sleeping to remove user input from:
// http://www.codeproject.com/system/display_states.asp
// The rest is all in MSDN.

// We create our own window to fire messages at.
// Older versions, and a lot of example code on the net, send to HWND_BROADCAST instead but this is bad:
// http://blogs.msdn.com/oldnewthing/archive/2006/06/13/629451.aspx

void runChildProcess(HINSTANCE hInstance, const TCHAR *szOffArg);

int _tmain(int argc, _TCHAR* argv[])
{
	LPCTSTR szClassNameMain = _T("ScreenSaveMain");

	HINSTANCE hInstance = GetModuleHandle(NULL);

	if (hInstance == NULL)
	{
		_tprintf(_T("\nScreenSave: Could not get HINSTANCE.\n"));
		return -1;
	}

	HANDLE hSleepEvent = CreateEvent(NULL, FALSE, FALSE, _T("ScreenSave_Sleep_{CB6E7615-A2EF-43da-8431-446AFE3F3368}"));

	if (hSleepEvent == NULL)
	{
		_tprintf(_T("\nScreenSave: Could not create sleep event.\n"));
		return -1;
	}

	HANDLE hPowerOffEvent = CreateEvent(NULL, FALSE, FALSE, _T("ScreenSave_Power_{CB6E7615-A2EF-43da-8431-446AFE3F3368}"));

	if (hPowerOffEvent == NULL)
	{
		_tprintf(_T("\nScreenSave: Could not create power-off event.\n"));
		CloseHandle(hSleepEvent);
		CloseHandle(hPowerOffEvent);
		return -1;
	}

	WNDCLASSEX wcex;
	wcex.cbSize = sizeof(WNDCLASSEX); 
	wcex.style			= 0; //CS_DBLCLKS | CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc	= DefWindowProc;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= 0;
	wcex.hInstance		= hInstance;
	wcex.hIcon			= NULL;
	wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
	wcex.lpszMenuName	= NULL;
	wcex.lpszClassName	= szClassNameMain;
	wcex.hIconSm		= LoadIcon(wcex.hInstance, (LPCTSTR)IDI_MAIN);

	ATOM atomMain = RegisterClassEx(&wcex);

	if (atomMain == 0)
	{
		_tprintf(_T("\nScreenSave: Could not register window class.\n"));
		CloseHandle(hSleepEvent);
		CloseHandle(hPowerOffEvent);
		return -1;
	}

	HWND hWndMain = CreateWindow(szClassNameMain, szClassNameMain, WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, HWND_MESSAGE, NULL, hInstance, NULL);

	if (hWndMain == NULL)
	{
		_tprintf(_T("\nScreenSave: Could not create window.\n"));
		CloseHandle(hSleepEvent);
		CloseHandle(hPowerOffEvent);
		return -1;
	}

	const TCHAR *szReadyEventName = NULL;

	bool bPreventSleep = false;
	bool bPreventPowerOff = false;

	bool bShowHelp = (argc <= 1);

	for (int i = 1; i < argc; ++i)
	{
		if (0 == _tcsicmp(argv[i], _T("/delay")))
		{
			DWORD dwDelayTime = 10000;
			if ((i+1) < argc && _T('/') != argv[i+1][0])
			{
				int iSecs = _ttoi(argv[++i]);

				if (iSecs >= 0)
				{
					dwDelayTime = iSecs;
					dwDelayTime *= 1000;
				}
				else
				{
					dwDelayTime = 0;
				}
			}

			Sleep(dwDelayTime);
		}
		else if (0 == _tcsicmp(argv[i], _T("/screensaver")))
		{
			SendMessage(hWndMain, WM_SYSCOMMAND, SC_SCREENSAVE, 0);
		}
		else if (0 == _tcsicmp(argv[i], _T("/lowpower")))
		{
			Sleep(500); // Stops recent user input cancelling the request.
			SendMessage(hWndMain, WM_SYSCOMMAND, SC_MONITORPOWER, 1);
		}
		else if (0 == _tcsicmp(argv[i], _T("/poweroff")))
		{
			Sleep(500); // Stops recent user input cancelling the request.
			SendMessage(hWndMain, WM_SYSCOMMAND, SC_MONITORPOWER, 2);
		}
		else if (0 == _tcsicmp(argv[i], _T("/disablesaver")))
		{
		   SystemParametersInfo(SPI_SETSCREENSAVEACTIVE, FALSE, NULL, SPIF_SENDCHANGE);
		}
		else if (0 == _tcsicmp(argv[i], _T("/enablesaver")))
		{
		   SystemParametersInfo(SPI_SETSCREENSAVEACTIVE, TRUE, NULL, SPIF_SENDCHANGE);
		}
		else if (0 == _tcsicmp(argv[i], _T("/disablesleep")))
		{
			runChildProcess(hInstance, _T("/preventsleep"));
		}
		else if (0 == _tcsicmp(argv[i], _T("/disablepoweroff")))
		{
			runChildProcess(hInstance, _T("/preventpoweroff"));
		}
		else if (0 == _tcsicmp(argv[i], _T("/enablesleep")))
		{
			if (!PulseEvent(hSleepEvent)) // It's an auto-reset event so this will wake one waiting process, allowing for nested disables.
			{
				_tprintf(_T("\nScreenSave: Failed to set sleep event.\n"));
			}
		}
		else if (0 == _tcsicmp(argv[i], _T("/enablepoweroff")))
		{
			if (!PulseEvent(hPowerOffEvent)) // It's an auto-reset event so this will wake one waiting process, allowing for nested disables.
			{
				_tprintf(_T("\nScreenSave: Failed to set power-off event.\n"));
			}
		}
		else if (0 == _tcsicmp(argv[i], _T("/preventsleep")))
		{
			bPreventSleep = true;
		}
		else if (0 == _tcsicmp(argv[i], _T("/preventpoweroff")))
		{
			bPreventPowerOff = true;
		}
		else if (0 == _tcsicmp(argv[i], _T("/readyeventname")) && (i+1) < argc && _T('/') != argv[i+1][0])
		{
			szReadyEventName = argv[++i];
		}
		else if (0 == _tcsicmp(argv[i], _T("/?"))
		||		 0 == _tcsicmp(argv[i], _T("-h")))
		{
			bShowHelp = true;
			break;
		}
		else
		{
			_tprintf(_T("\nScreenSave: Invalid argument: \"%s\"\nRun \"ScreenSave.exe /?\" for help\n"), argv[i]);
		}
	}

	PostQuitMessage(0);

	MSG msg;
	BOOL bRet;
	while (bRet = GetMessage(&msg, NULL, 0, 0)) 
	{
		if (-1 == bRet)
		{
			break;
		}

	//	if (!TranslateAccelerator(msg.hwnd, hAccelTable, &msg)) 
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	// We do this after closing the window so we don't leave an unpumped window around while we wait.

	if (bPreventSleep && bPreventPowerOff)
	{
		_tprintf(_T("\nScreenSave: The internal /PreventSleep and /PreventPowerOff args shouldn't be used together.\nUse the disable* arguments rather than the prevent* ones.\n"));
	}
	else if (bPreventSleep || bPreventPowerOff)
	{
		bool bWait = true;

		if (NULL == SetThreadExecutionState(
				ES_CONTINUOUS
			    | (bPreventSleep ? ES_SYSTEM_REQUIRED : 0)
				| (bPreventPowerOff ? ES_DISPLAY_REQUIRED : 0)))
		{
			_tprintf(_T("\nScreenSave: SetThreadExecutionState failed.\n"));
			bWait = false;
		}

		if (bWait && szReadyEventName != NULL)
		{
			HANDLE hReadyEvent = OpenEvent(EVENT_MODIFY_STATE, FALSE, szReadyEventName);
			
			if (hReadyEvent == NULL)
			{
				_tprintf(_T("\nScreenSave: Failed to open the ready event.\n"));
				bWait = false;
			}
			else
			{
				if (!SetEvent(hReadyEvent))
				{
					_tprintf(_T("\nScreenSave: Failed to set ready event.\n"));
					bWait = false;
				}

				CloseHandle(hReadyEvent);
			}
		}

		if (bWait)
		{
			WaitForSingleObject(bPreventSleep ? hSleepEvent : hPowerOffEvent, INFINITE);
		}
	}

	if (bShowHelp)
	{
		_tprintf(_T("\nUsage:\n"));
		_tprintf(_T("\n       ScreenSave [/delay [number of seconds]] [/screensaver] [/lowpower] [/poweroff]"));
		_tprintf(_T("\n\n/delay defaults to 10 seconds if no number given."));
		_tprintf(_T("\n\nYou can give multiple /delay arguments. For example, this will wait 10 sec"));
		_tprintf(_T("\nthen run the screen saver, wait another 20 sec, then turn off the monitor:\n"));
		_tprintf(_T("\n       ScreenSave /delay 10 /screensaver /delay 20 /poweroff"));

		_tprintf(_T("\n\nMonitor power-off will not lock your workstation by itself. If your"));
		_tprintf(_T("\nscreensaver is configured to lock the computer then you can use the"));
		_tprintf(_T("\ncommand below to immediately turn your monitors off and lock the machine:\n"));
		_tprintf(_T("\n       ScreenSave /screensaver /poweroff"));
		_tprintf(_T("\n\nYou can also enable/disable the screensaver, monitor power-off\n"));
		_tprintf(_T("and system sleep/hibernation:\n"));
		_tprintf(_T("\n       ScreenSave /disableSaver"));
		_tprintf(_T("\n       ScreenSave /enableSaver"));
		_tprintf(_T("\n       ScreenSave /disablePowerOff"));
		_tprintf(_T("\n       ScreenSave /enablePowerOff"));
		_tprintf(_T("\n       ScreenSave /disableSleep"));
		_tprintf(_T("\n       ScreenSave /enableSleep"));
		_tprintf(_T("\n\nThose options can be combined. For example:\n"));
		_tprintf(_T("\n       ScreenSave /disableSaver /disablePowerOff /disableSleep"));
		_tprintf(_T("\n\nIf you disable sleep/hibernation with /disableSleep then a copy of"));
		_tprintf(_T("\nScreenSave.exe will run in the background until you enable it again with"));
		_tprintf(_T("\n/enableSleep. The background process will consume virtually no memory or CPU."));
		_tprintf(_T("\nThis is similarly true if you disable monitor power-off."));
		_tprintf(_T("\n\nCalls to /disablePowerOff and /disableSleep will nest. You must call the"));
		_tprintf(_T("\nrespective /enable* commands an equal number of times to undo them."));
		_tprintf(_T("\nOn the other hand, calls to /disableSaver do not nest."));
		_tprintf(_T("\n\n"));
	}

	CloseHandle(hSleepEvent);
	CloseHandle(hPowerOffEvent);
	return 0;
}

void runChildProcess(HINSTANCE hInstance, const TCHAR *szOffArg)
{
	TCHAR szChildReadyEventName[128];
	_stprintf_s(szChildReadyEventName, _T("ScreenSave_Ready_%lx_{CB6E7615-A2EF-43da-8431-446AFE3F3368}"), GetCurrentProcessId());

	HANDLE hChildReadyEvent = CreateEvent(NULL, TRUE, FALSE, szChildReadyEventName);

	if (hChildReadyEvent == NULL)
	{
		_tprintf(_T("\nScreenSave: Failed to create child ready event.\n"));
	}
	else
	{
		TCHAR szOwnPath[_MAX_PATH];

		if (!GetModuleFileName(hInstance, szOwnPath, _countof(szOwnPath)))
		{
			_tprintf(_T("\nScreenSave: Failed to get own exe path.\n"));
		}
		else
		{
			LPWCH envStringsW = GetEnvironmentStringsW();

			STARTUPINFO startupInfo = {0};
			startupInfo.cb = sizeof(startupInfo);
			PROCESS_INFORMATION processInfo = {0};

			TCHAR szChildCommandLine[1024];
			_stprintf_s(szChildCommandLine, _T("ScreenSave.exe %s /readyeventname %s"), szOffArg, szChildReadyEventName);

			if (!CreateProcess(szOwnPath, szChildCommandLine, NULL, NULL, FALSE,
				               CREATE_UNICODE_ENVIRONMENT, envStringsW, NULL,
							   &startupInfo, &processInfo))
			{
				_tprintf(_T("\nScreenSave: Failed to create child process.\n"));
			}
			else
			{
				HANDLE childWaitHandles[2];
				childWaitHandles[0] = hChildReadyEvent;
				childWaitHandles[1] = processInfo.hProcess;

				if (WAIT_OBJECT_0 != WaitForMultipleObjects(_countof(childWaitHandles), childWaitHandles, FALSE, INFINITE))
				{
					_tprintf(_T("\nScreenSave: Child process failed.\n"));
				}

				CloseHandle(processInfo.hProcess);
				CloseHandle(processInfo.hThread);
			}


			if (envStringsW)
			{
				FreeEnvironmentStringsW(envStringsW);
			}
		}

		CloseHandle(hChildReadyEvent);
	}
}
